#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-dolcenum-py
"""Enum metaclass to assign member docstrings to the members.

The idea was found at https://stackoverflow.com/a/78943193.
"""

import enum
import inspect
from typing import Self, Type


class MissingDocstringError(Exception):
    """DoclEnumMeta-derived enum is missing one or more docstrings."""
    pass


class DolcEnumMeta(enum.EnumMeta):
    """Metaclass to assign member docstrings to the enum members.

    The metaclass is assigned to the `metaclass` attribute in the enum class
    definition. The class itself and each member **must** have a docstring
    (or there must be **no** docstrings whatsoever.)

    :raise ValueError: One or more docstrings are missing.

    Example
    -------

    .. code-block:: python

        import enum
        from pymisclib.dolcenum import DolcEnumMeta


        @enum.unique
        class MyEnum(enum.IntEnum, metaclass=DolcEnumMeta):
            \"\"\"Class MyEnum docstring.\"\"\"
            A = enum.auto()
            \"\"\"Member MyEnum.A docstring.\"\"\"
            B = enum.auto()
            \"\"\"Member MyEnum.B docstring.\"\"\"
            C = enum.auto()
            \"\"\"Member MyEnum.C docstring.\"\"\"


        for e in MyEnum:
            print(f'{e.name}:{e.value} "{e.__doc__}"')
    """
    def __new__(metaclass: Type[Self], name: str, bases: tuple[type, ...], attrs: enum._EnumDict):
        """Create a new instance with docstrings for each member."""
        cls = super().__new__(metaclass, name, bases, attrs)

        # Get the source code and split comment into individual docstrings.
        source = inspect.getsource(cls)
        docstrings = source.split('"""')[-2::-2]

        if len(docstrings) == 0:
            # No docstrings means nothing to do.
            return cls
        # Check for consistency.
        if len(docstrings) < len(cls.__members__) + 1:
            raise MissingDocstringError('missing docstring')

        # Assign a docstring to each member.
        for member_name, member_docstring in zip(reversed(cls.__members__), docstrings):
            enum_member = getattr(cls, member_name, None)
            enum_member.__doc__ = member_docstring.strip()

        return cls


def example():

    @enum.unique
    class Enum1(enum.Enum, metaclass=DolcEnumMeta):
        """Enum1 class docstring."""
        E1 = enum.auto()
        """Enum1.E1 member docstring."""
        E2 = enum.auto()
        """Enum1.E2 member docstring."""
        E3 = enum.auto()
        """Enum1.E3 member docstring."""

    print('Enum1 has a docstring per member.')
    for e in Enum1:
        print(f'    {e.name}:{e.value} "{e.__doc__}"')

    try:
        @enum.unique
        class Enum2(enum.Enum, metaclass=DolcEnumMeta):
            """Enum2 class docstring."""
            E1 = enum.auto()
            """Enum2.E1 member docstring."""
            E2 = enum.auto()
            """Enum2.E2 member docstring."""
            E3 = enum.auto()

        print('Enum2 is missing the last docstring.')
        for e in Enum2:
            print(f'    {e.name}: {e.__doc__}')
    except MissingDocstringError as e:
        print(f'Missing docstring detected: {e}')


if __name__ == '__main__':
    example()




#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023, 2024, 2025 by Christian Dönges
# SPDXID: SPDXRef-utilities-py
"""Collection of various utility functions.
"""

# pylint: disable=invalid-name

import argparse
import bz2
import codecs
import contextlib
import ctypes
import glob
import gzip
import locale
import logging
import lzma
import os
import platform
import re
import shutil
import sys
import traceback
from datetime import datetime, timedelta, timezone, tzinfo, UTC
from pathlib import Path
from pprint import pformat
from time import mktime
from typing import Any, Generator
from zipfile import ZipFile, ZipInfo


from pymisclib.xmlutils import log_xml


# Globals (use sparingly)
logger = logging.getLogger(__name__)


def convert_dotnet_datetime_ticks_to_datetime(ticks: int,
                                              src_tz: tzinfo|None = None) -> datetime:
    """Convert .net System.DateTime.ticks to a Python datetime.

    A .net tick is 100 ns since 01-JAN-0001 00:00:00 in the local timezone.

    :param int ticks: Number of ticks to convert.
    :param tzinfo|None src_tz: Timezone of the ticks or None to
        create a timezone naive datetime object (i.e. effectively localtime).
    :returns: Python datetime.
    :raises ValueError: The ticks can not be converted to a Python datetime.

    .. versionadded:: 1.6.1
    """
    seconds = ticks / 10**7
    milliseconds = ticks / 10**4 - seconds * 1000
    epoch = datetime(1, 1, 1, tzinfo=src_tz)
    td = timedelta(seconds=seconds, milliseconds=milliseconds)
    dt = epoch + td
    return dt


def convert_s_since_epoch_to_datetime(seconds: int|float) -> datetime:
    """Convert seconds since the UNIX epoch to a Python datetime.

    The UNIX expoch starts at 01-JAN-1970T00:00:00 UTC.

    :param int|float seconds: Number of seconds since the UNIX epoch in UTC.
        If this is a floating-point number, the fractiona seconds are added
        to the datetime, increasing precision up to microseconds (which is
        the Python implementation limit).
    :returns: Python datetime.
    :rtype: datetime
    :raises ValueError: The seconds can not be converted to a Python datetime.

    .. versionadded:: 1.7.0
    """
    ts = datetime.fromtimestamp(seconds, tz=UTC)
    if int(seconds) != seconds:
        fractional_seconds = seconds - int(seconds)
        td = timedelta(seconds=fractional_seconds)
        ts = ts + td
    return ts


def convert_ms_since_epoch_to_datetime(milliseconds: int) -> datetime:
    """Convert milliseconds since the UNIX epoch to a Python datetime.

    The UNIX expoch starts at 01-JAN-1970T00:00:00 UTC.

    :param int milliseconds: Number of milliseconds since the UNIX epoch in UTC.
    :returns: Python datetime.
    :rtype: datetime
    :raises ValueError: The milliseconds can not be converted to a Python datetime.

    .. versionadded:: 1.6.1
    """
    return convert_us_since_epoch_to_datetime(milliseconds * 1000)


def convert_us_since_epoch_to_datetime(microseconds: int) -> datetime:
    """Convert microseconds since the UNIX epoch to a Python datetime.

    The UNIX expoch starts at 01-JAN-1970T00:00:00 UTC.

    :param int microseconds: Number of microseconds since the UNIX epoch in UTC.
    :returns: Python datetime with timezone UTC.
    :rtype: datetime
    :raises ValueError: The microseconds can not be converted to a Python datetime.

    .. seealso::

        `time_t <https://en.cppreference.com/w/c/chrono/time_t>`_
            Although not defined by the C standard, this is almost always an
            integral value holding the number of seconds (not counting leap
            seconds) since 00:00, Jan 1 1970 UTC, corresponding to POSIX time.

        `Unix time <https://en.wikipedia.org/wiki/Unix_time>`_
            Unix time is currently defined as the number of non-leap seconds
            which have passed since 00:00:00 UTC on Thursday, 1 January 1970,
            which is referred to as the Unix epoch.

    .. versionadded:: 1.7.0
    """
    # Split the time into seconds and remaining microseconds.
    time_s = int(microseconds / 1000000)
    time_us = microseconds - (time_s * 1000000)
    # Create a datetime object with seconds precision.
    ts_s = datetime.fromtimestamp(time_s, tz=UTC)
    # Use the seconds precision object to create a microsecond precision
    # datetime object.
    ts = datetime(ts_s.year, ts_s.month, ts_s.day,
                  ts_s.hour, ts_s.minute, ts_s.second,
                  time_us, tzinfo=UTC)
    return ts


def convert_timestamp_to_datetime(timestamp: int|float,
                                  fmt: str = 's',
                                  src_tz: tzinfo | None = None) -> datetime:
    """Convert a timestamp to a Python datetime.

    :param int|float timestamp: Timestamp to convert.
    :param str fmt: Format of the timestamp. Any one of 's', 'ms', 'µs', 'us',
         or '.net'.
    :param tzinfo|None src_tz: Timezone of the ticks or None to
        use the configured default timezone (i.e. localtime). **This parameter
        is ignored unless fmt is set to '.net'**.
    :returns: Python datetime.
    :rtype: datetime
    :raises ValueError: The format is not supported or the timestamp can not
         be converted to a Python datetime.
    """
    if fmt == 's':
        return convert_s_since_epoch_to_datetime(timestamp)
    elif fmt == 'ms':
        return convert_ms_since_epoch_to_datetime(int(timestamp))
    elif (fmt == 'µs') or (fmt == 'us'):
        return convert_us_since_epoch_to_datetime(int(timestamp))
    elif fmt == '.net':
        return convert_dotnet_datetime_ticks_to_datetime(int(timestamp), src_tz)
    raise ValueError(f'Unknown timestamp format "{fmt}"')


def dir_path(path: str) -> Path:
    """Convert the string to a path and return it if it is a directory.

        This function is intended to be used as the `type` argument to :func:`Argparser.add_argument`.

        :param str path: String containing path to check.
        :return: Path to a directory.
        :rtype: Path

        :raise argparse.ArgumentTypeError: The string was not a valid path to a directory.
    """
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError(
            f'"{path}" is not a valid path to a directory')
    return Path(path)


def exit_hard(returncode: int = 0):
    """Terminate the running application.

        :param int returncode: Exit code to return to spawning shell.
    """
    os._exit(returncode)  # pylint: disable=protected-access


def extract_from_zip_preserving_mtime(zip_file: ZipFile,
                                      zip_info: ZipInfo,
                                      out_path: dir_path):
    """Extract file from a ZIP archive preserving the modification time.

        :param ZipFile zip_file: The archive to extract from.
        :param ZipInfo zip_info: The information about the file to extract.
        :param dir_path out_path: Destination directory for the extracted
            file (must exist).

        :see: https://stackoverflow.com/q/9813243
    """
    zip_file.extract(zip_info, out_path)
    # Add (0, 0, -1) to the time to indicate we do not know if DST was active.
    date_time = mktime(zip_info.date_time + (0, 0, -1))
    # Set ctime and mtime of the extracted file.
    os.utime(out_path / zip_info.filename, (date_time, date_time))


def file_path(path: str) -> Path:
    """Convert the string to a path and return it if it is a file.

        :param str path: String containing path to check.
        :return: Path to a file.
        :rtype: Path

        :raise argparse.ArgumentTypeError: The string did not contain a valid
            path to a file.
    """
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError(
            f'"{path}" is not a valid path to a file')
    return Path(path)


def file_birthtime(path: os.PathLike,
                   tz: timezone|None = None,
                   fudge: bool = True) -> datetime:
    """Return a file system object birthtime.

    .. note::
        If a platform/file system (like Linux with EXT4) does not support
        `st_birthtime`, try `fudge=True` to get an approximation. This
        approximation can be way off the birthtime of the file system object.

    :param os.PathLike path: Path to a filesystem object (e.g. file or directory.)
    :param tz|None: Timezone used by the filesystem.
    :param bool fudge: If `False`, use only `st_birthtime` and raise an
        AttributeError if this field is unsupported. If `True`, try
        `st_birthtime`, `st_ctime`, and `st_mtime` in that order.
    :returns: Birthtime of the file.
    :rtype: datetime
    :raises FileNotFoundError: The given path does not exist.
    :raises AttributeError: Unable to determine creation time.
    """
    st = os.stat(path)

    # According to the Python documentation, st_birthtime is canonical.
    # Some OSes still use st_ctime for birthtime.
    # If all else fails, fall back to modification time.
    if fudge:
        fields = ['st_birthtime', 'st_ctime', 'st_mtime']
    else:
        fields = ['st_birthtime']
    for field in fields:
        ts = getattr(st, field, None)
        if ts is not None:
            return datetime.fromtimestamp(ts, tz=tz)

    raise AttributeError('birthtime not supported')


def get_language() -> str:
    """Determine the language the current user has set their OS to."""
    if os.name == 'posix':
        # BSD, Darwin, and Linux make it easy.
        lang = os.environ['LANG'].split('.')[0]
    elif os.name == 'nt':
        windll = ctypes.windll.kernel32
        lang = locale.windows_locale[windll.GetUserDefaultUILanguage()]
    else:
        raise RuntimeError(f'unknown OS: {os.name}')
    return lang


def hexdump(b: bytes,
            bytes_per_line: int = 16,
            start_offset: int = 0,
            offset_digits: int = 8,
            show_ascii: bool = True) -> Generator[str, None, None]:
    """Generator function to create a pretty representation of the given bytes and return a line per call.

        ```
        first_prefix   00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        always_prefix  00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        ```

        :param bytes b: The bytes to print.
        :param int bytes_per_line: Number of bytes per line.
        :param int start_offset:   Starting offset for the first byte.
        :param int offset_digits:  Number of digits in the offset.
        :param bool show_ascii: Print ASCII characters after the hex dump if True.
        :return: A single hexdump line per call. None when done.
        :rtype: Generator[str, None, None]
    """
    length = len(b)
    offset = 0
    if length <= bytes_per_line:
        show_offset = False
    else:
        show_offset = True
    while length > 0:
        if length < bytes_per_line:
            bytes_per_line = length
        b_slice = b[offset:offset + bytes_per_line]
        hs = b_slice.hex(' ')
        if show_offset:
            s = f'{offset + start_offset:0{offset_digits}x}  {hs}  '
        else:
            s = f'{hs}  '
        if show_ascii:
            for c in b_slice:
                c = chr(c)
                if c.isprintable():
                    s += c
                else:
                    s += '.'
        length -= bytes_per_line
        offset += bytes_per_line
        yield s


def humansort(unsorted: list[str],
              ignore_case: bool = False) -> list[str]:
    """Perform human-like sorting of a list of strings.

    :param list[str] unsorted: A list of strings to sort.
    :param bool ignore_case: Sort ignoring case if `True`.
      Default is `False` with upper case sorted before lower case of the same
      letter.
    :return: A new list of sorted strings.
    :rtype: list[str]

    Given a list of strings, a new list is sorted numerically and
    alphabetically instead of just alphabetically
    (like Python :py:func:`sorted`).

    * Pure number (float or int) are sorted numerically.
    * Pure alpha strings are sorted alphabetically.
    * Strings containing an alpha followed by a number component are sorted
      alphabetically, then numerically.
    * Strings containing a number followed by an alpha are sorted
      numerically, then alphabetically.

    .. note::
      float and int zeros are sorted as zero with no regard for
      the number of digits.

      .. code-block:: python

        humansort(['0.00', '0', '0.0'])
        ['0.00', '0', '0.0']

    .. code-block:: python

      humansort(['Z', 'z', '1.0', '-5', '2z', '1z', 'Z2.1', 'Z1',
                 'Z2', '1a2', '1a1', '1a2b2', '1a2b1', '1a1b1',
                 'z1', 'z1a'], ignore_case=True)
      ['-5', '1.0', '1a1', '1a1b1', '1a2', '1a2b1', '1a2b2', '1z',
       '2z', 'Z', 'z', 'Z1', 'z1', 'z1a', 'Z2', 'Z2.1']
    """
    def to_tuple(s: str) -> tuple[str, int]:
        """Transform the given string into a tuple."""
        output = ()
        s_split = re.split(r'([-+]?[0-9]*\.?[0-9]*)', s)
        string = ''
        default_number = float('-inf')
        for m in s_split:
            if len(m) == 0:
                # Skip empty matches.
                continue
            try:
                number = float(m)
                if len(string) > 0:
                    output += (string, default_number)
                    string = ''
                output += ('', number)
            except ValueError:
                if ignore_case:
                    string += m.lower()
                else:
                    string += m

        if len(string) > 0:
            output += (string, default_number)
        return output
    return sorted(unsorted, key=to_tuple)


def initialize_console():
    """Initialize the console to work with UTF-8 encoded strings.

        On windows, the console is strange and if output is redirected to a
        file, it gets even stranger. This confuses Python and even though
        PEP 528 solves the problem for interactive consoles, this does not
        help for non-interactive (which redirected streams are).

        The solution for now is to reconfigure the codecs for stdout and
        stderr to always use UTF-8 and replace unmappable characters.
    """
    if os.name == 'nt':
        if sys.stdout.encoding != 'utf-8':
            sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'replace')
            print('Reconfigured stdout to use utf-8 encoding.')
        if sys.stderr.encoding != 'utf-8':
            sys.stderr = codecs.getwriter('utf-8')(sys.stderr.buffer, 'replace')
            print('Reconfigured stderr to use utf-8 encoding.')


def iso8601_str(d: datetime,
                seconds_digits: int = 0,
                compact: bool = False,
                utc: bool = False) -> str:
    """Convert a datetime to an ISO8601 string.

        An ISO8601 compliant string is not fully defined, implementations
        have some leeway in the formatting. This function will produce various
        strings, depending on the input parameters.

        The timezone information ("±HHMM") is only present for aware datetime
        objects (e.g. datetime with tzinfo).

        The number of fractional seconds digits is given by seconds_digits. If
        zero digits are requested, the decimal point is left out (e.g.
        "2023-11-21T18:37:31"). If more than 6 digits are requested, only 6
        digits will be output because datetime objects are limited to
        microsecond resolution.

        Compact notation does not separate the different fields of the string
        (e.g. "YYYY-MM-DDTHH:MM:SS" becomes "YYYYMMDDTHHMMSS").

        UTC notation will convert the datetime into UTC. Note that if the
        datetime is not aware, Python will assume it is in localtime. Instead
        of the time zone offset, a "Z" will be appended.

        :param datetime d: Date to convert to a string.
        :param int seconds_digits: Digit of precision of the seconds.
        :param bool compact: Use separators between values if False,
            compact representation if True.
        :param bool utc: Output UTC time with 'Z' suffix instead of timezone offset.
        :return: String representation of the datetime.

        **Example**

        Given the datetime 2023-11-21T18:37:31.123456+0100 and that we are
        running in GMT-02, the following will be output:

        ===== ============== ======= ===== ===============================
        aware seconds_digits compact utc   Result
        ===== ============== ======= ===== ===============================
        Yes   0              False   False 2023-11-21T18:37:31+0100
        Yes   1              False   False 2023-11-21T18:37:31.1+0100
        Yes   6              False   False 2023-11-21T18:37:31.123456+0100
        Yes   7              False   False 2023-11-21T18:37:31.123456+0100
        No    6              False   False 2023-11-21T18:37:31.123456
        Yes   6              True    False 20231121T183731.123456+0100
        Yes   6              False   True  2023-11-21T17:37:31.123456Z
        No    6              False   True  2023-11-21T15:37:31.123456Z
        ===== ============== ======= ===== ===============================

        **Note**: non-aware and utc=True result depends on current timezone.

        .. versionadded:: 1.4.1
    """
    if utc:
        d = d.astimezone(tz=timezone.utc)
    if compact:
        s = f'{d.strftime("%Y%m%dT%H%M%S")}'
    else:
        s = f'{d.strftime("%Y-%m-%dT%H:%M:%S")}'
    if seconds_digits > 0:
        if seconds_digits > 6:
            seconds_digits = 6  # datetime is only precise to microseconds
        fractional_seconds = int(d.microsecond / 10**(6-seconds_digits))
        s += f'.{fractional_seconds:0{seconds_digits}d}'
    if utc:
        s += 'Z'
    elif d.tzinfo:
        s += f'{d.strftime("%z")}'
    return s


def logging_add_trace():
    """Add a loglevel TRACE.

        This function should be called only once.
    """
    if hasattr(logging_add_trace, "nr_calls"):
        logging_add_trace.nr_calls += 1
        logging.getLogger(__name__).warning(
            'logging_add_trace() called %d times.',
            logging_add_trace.nr_calls)
        return
    logging_add_trace.nr_calls = 1  # it doesn't exist yet, so initialize it

    if hasattr(logging, 'TRACE'):
        logging.getLogger(__name__).debug('TRACE logging already enabled.')

    # Perform function.
    logging.TRACE = 9
    logging.addLevelName(logging.TRACE, 'TRACE')

    def trace(self, message, *args, **kws):
        """Output message with level TRACE."""
        if self.isEnabledFor(logging.TRACE):
            # Logger takes its '*args' as 'args'.
            self._log(logging.TRACE, message, args, **kws)  # pylint: disable=protected-access

    logging.Logger.trace = trace


def logging_initialize(
        loglevel: int = logging.WARNING,
        log_dir_path: Path = Path('.'),
        log_file_name_format: str = '%P.log',
        log_rotation: int = 0,
        log_compression: int = 5,
        loglevel_console: int = logging.WARNING,
        loglevel_file: int = logging.DEBUG) -> logging.Logger:
    """Initialize logging and return an instance of the root logger.

        If a submodule uses :py:`logging.getLogger('somename')`, the logger will
        be a child of the root logger and inherit the settings made here.

        :param int loglevel: Loglevel of the logger. Log messages not meeting
            the level requirement are not logged.
        :param Path log_dir_path: Path of the directory containing log files.
            :py:`None` will prevent log file creation.
        :param str log_file_name_format: Format string for the log file name.
            :py:`None` will prevent log file creation.
        :param int log_rotation: How many logfiles of the same name to keep.
            If set to 0, any existing log file is overwritten. If set to >0,
            that many old log file copies (named <name>.1, <name>.2, etc.)
            are retained.
        :param int log_compression: Zlib compression level to use in compressing
            log. Level 0 is no compression, level 9 is the highest possible.
        :param int loglevel_console: Loglevel filter of the console logger.
        :param int loglevel_file: Loglevel filter of the file logger.
        :return: The root logger instance for the application.
        :rtype: logging.logger

        :note: If the :py:`log_file_name_format` contains a timestamp, log_rotation
            will only work on other log file copies with the exact same timestamp.
    """
    # Configure the root logger instance.
    global logger   # # pylint: disable=global-statement
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # Configure the console handler.
    lc = locale.getpreferredencoding()
    if lc.lower() != 'utf-8':
        locale.setlocale(locale.LC_CTYPE, 'C')
    initialize_console()
    ch = logging.StreamHandler()
    ch.setLevel(loglevel_console)
    terse_formatter = logging.Formatter('%(levelname)-8s %(message)s')
    ch.setFormatter(terse_formatter)
    logger.addHandler(ch)
    logger.debug('Logging to console initialized.')

    if log_file_name_format is not None and log_dir_path is not None:
        log_file_name = string_from_format(log_file_name_format)
        log_file_path = log_dir_path / log_file_name
        rotate_file(log_file_name, log_dir_path, log_rotation, log_compression)

        fh = logging.FileHandler(log_file_path, encoding='utf-8', mode='w')
        fh.setLevel(loglevel_file)
        verbose_formatter = logging.Formatter('%(asctime)s %(name)-26s %(levelname)-8s %(message)s')
        fh.setFormatter(verbose_formatter)
        logger.addHandler(fh)
        logger.debug('Logging to file "%s" initialized.', log_file_path)

    if lc.lower() != 'utf-8':
        if locale.getpreferredencoding().lower() == 'utf-8':
            logger.debug('Changed encoding from "%s" to "utf-8".', lc)
        elif lc == locale.getpreferredencoding():
            logger.debug('Failed to change encoding from "%s" to "utf-8".', lc)
        else:
            logger.warning('Failed to change encoding from "%s" to "utf-8", got "%s".',
                           lc, locale.getpreferredencoding())

    return logger


def initialize_logging(args: argparse.Namespace):
    """Initialize the logging interface with the command line options
        passed through the object 'args'. An instance of the root logger is
        returned to the caller.

        If a submodule uses `logging.getLogger('somename')`, the logger will
        be a child of the root logger and inherit the settings made here.

        :param argparse.Namespace args: Namespace containing attributes used as parameters.
        :return: The root logger instance for the application.
        :rtype: logging.logger

        Namespace attributes used:
            * `args.debug` bool - Enable or disable debug mode.
            * `args.logging` str - Loglevel (e.g. CRITICAL, ERROR, etc.)
            * `args.verbose` bool - Enable or disable verbose mode.

        .. deprecated:: 1.2.0
           Use :func:`logging_initialize` and :func:`logging_add_trace()` (if
           needed) instead.
    """

    # Define a new log level 'trace'
    logging_add_trace()

    # args.loglevel contains the string value of the command line option
    # --loglevel.
    numeric_log_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(numeric_log_level, int):
        raise ValueError(f'Invalid log level: {args.loglevel}')

    # Configure the root logger instance.
    global logger   # # pylint: disable=global-statement
    logger = logging.getLogger()
    if args.debug:
        logger.setLevel(logging.TRACE)
    else:
        logger.setLevel(numeric_log_level)

    # Create file handler only if debug mode is active.
    if args.debug:
        app_name = sys.argv[0]
        fh = logging.FileHandler(app_name + '.log', encoding='utf-8', mode='w')
        fh.setLevel(logging.TRACE)

    lc = locale.getpreferredencoding()
    if lc != 'utf-8':
        locale.setlocale(locale.LC_CTYPE, 'C')

    # Create a console handler.
    initialize_console()
    ch = logging.StreamHandler()
    ch.setLevel(numeric_log_level)

    # Create a formatter and add it to the handlers.
    terse_formatter = logging.Formatter('%(levelname)-8s %(message)s')
    ch.setFormatter(terse_formatter)
    if args.debug:
        verbose_formatter = logging.Formatter(
            '%(asctime)s %(name)-26s %(levelname)-8s %(message)s')
        fh.setFormatter(verbose_formatter)

    # Add the handlers to the logger.
    logger.addHandler(ch)
    if args.debug:
        logger.addHandler(fh)

    if args.verbose:
        logger.info('Logging initialized.')

    if lc != 'utf-8':
        if locale.getpreferredencoding() == 'utf-8':
            logger.debug('Changed encoding from "%s" to "utf-8".', lc)
        elif lc == locale.getpreferredencoding():
            logger.debug('Failed to change encoding from "%s" to "utf-8".', lc)
        else:
            logger.warning('Failed to change encoding from "%s" to "utf-8", got "%s".',
                           lc, locale.getpreferredencoding())

    logger.debug('Parsed args are: %s', args)
    logger.warning('utilities.initialize_logging() is deprecated, use utilities.logging_initialize() instead.')
    return logger


def is_power_of_two(n: int) -> bool:
    """Return True if the given number *n* is a power of two.

        :param int n: number to check
        :return: True if *n* is a power of two, False otherwise.
        :rtype: bool
    """
    return (n != 0) and ((n & (n - 1)) == 0)


def log_hexdump(fn_logger: logging.Logger,
                b: bytes,
                bytes_per_line: int = 16,
                level: int = logging.DEBUG,
                start_offset: int = 0,
                first_prefix: str = '',
                always_prefix: str = '',
                show_ascii: bool = True):
    """Log a pretty representation of the given bytes to the logger.

        ```
        first_prefix   00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        always_prefix  00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        ```

        :param logging.Logger fn_logger: The logger to log to.
        :param bytes b: The bytes to print.
        :param int bytes_per_line: The number of bytes per line.
        :param int level: Level for logging (e.g. CRITICAL, ERROR, .. DEBUG)
        :param int start_offset: The starting offset for the first byte.
        :param str first_prefix: A string before the offset on the first line.
        :param str always_prefix: A string that will be printed before every
            line (except the first if first_prefix was specified).
        :param bool show_ascii: Print ASCII characters after the hex dump if True.
    """
    if len(first_prefix) == 0:
        first_prefix = always_prefix
    elif len(first_prefix) > len(always_prefix):
        always_prefix += ' ' * (len(first_prefix) - len(always_prefix))
    elif len(always_prefix) > len(first_prefix):
        first_prefix += ' ' * (len(always_prefix) - len(first_prefix))
    first = True
    for line in hexdump(b,
                        bytes_per_line=bytes_per_line,
                        start_offset=start_offset,
                        offset_digits=8,
                        show_ascii=show_ascii):
        if first:
            fn_logger.log(level, '%s  %s', first_prefix, line)
            first = False
        else:
            fn_logger.log(level, '%s  %s', always_prefix, line)


def log_pp(fn_logger: logging.Logger,
           obj: Any,
           level: int = logging.DEBUG,
           indent: int = 0,
           nesting_indent: int = 1,
           max_width: int = 80,
           compact: bool = False,
           sort_dicts: bool = True):
    """Log an object formatted with pprint.pformat to a logger.
    Mutli-line output is spread to multiple log entries.

    :param logging.Logger fn_logger: Logger instance to log to.
    :param Any obj: Object to log.
    :param int level: Level for logging (e.g logging.DEBUG, logging.INFO, ...)
    :param int indent: Number of spaces to indent all entries.
    :param int nesting_indent: Additional indentation for each object nesting
        level.
    :param int max_width: Maximum width of the formatted object if `indent` is
        zero. The maximum log message length is `max_width + indent`.
    :param bool compact: True to use compact representation, False for pretty.
    :param bool sort_dicts: True to sort dictionaries before output.

    .. csv-table:: log_pp() Parameter correspondence to `pprint.PrettyPrinter <https://docs.python.org/3/library/pprint.html#pprint.PrettyPrinter>`_
        :header: "log_pp", "pprint.PrettyPrinter"
        :align: left

        "`nesting_indent`", "`indent`"
        "`max_width`", "`width`"
        "`compact`", "`compact`"
        "`sort_dicts`", "`sort_dicts`"
    """
    if not fn_logger.isEnabledFor(level):
        return
    fmt = ' ' * indent + '%s'
    for line in pformat(object=obj,
                        indent=nesting_indent,
                        width=max_width,
                        compact=compact,
                        sort_dicts=sort_dicts).splitlines():
        fn_logger.log(level, fmt, line)


def log_stacktrace(fn_logger: logging.Logger,
                   level: int = logging.DEBUG):
    """Log the current stack trace inside or outside an exception.

        :param logging.Logger fn_logger: The logger to log to.
        :param int level: Level for logging (e.g. CRITICAL, ERROR, .. DEBUG)
    """
    # Get the current exception.
    ex = sys.exc_info()[0]
    # Remove this method from the call stack.
    stack = traceback.extract_stack()[:-1]
    if ex is not None:
        # Exception is present, so remove the call of this method.
        del stack[-1]
    fn_logger.log(level, 'Traceback (most recent call last):')
    for s in traceback.format_list(stack):
        fn_logger.log(level, s)
    if ex is not None:
        fn_logger.log(level, traceback.format_exc())


def netmask_string_to_cidr(netmask: str) -> int:
    """Convert an IPv6 netmask string to a CIDR.

    :param str netmask: Netmask string to convert.
    :return: CIDR.
    :rtype: int

    Example
    -------
    .. code-block:: python

        netmask_string_to_cidr('ffff:fff0::')

    .. versionadded:: 1.5.3
    """

    # From https://stackoverflow.com/a/33533007/7083698
    bit_count = [
        0,
        0x8000,
        0xC000,
        0xE000,
        0xF000,
        0xF800,
        0xFC00,
        0xFE00,
        0xFF00,
        0xFF80,
        0xFFC0,
        0xFFE0,
        0xFFF0,
        0xFFF8,
        0xFFFC,
        0xFFFE,
        0xFFFF,
    ]  # bit_count

    count = 0
    for w in netmask.split(':'):
        if not w or int(w, 16) == 0:
            break
        count += bit_count.index(int(w, 16))

    return count


class OsInfo:
    """Detect the operating system the application is running on.
    """
    _bsd_variants = {"FreeBSD", "OpenBSD", "NetBSD", "DragonFly"}
    """Static variable containing known BSD variants."""
    def __init__(self):
        self._os_name = platform.system()
        self._os_platform = platform.platform()
        self._os_release = platform.release()
        self._os_system = sys.platform
        self._cpu_architecture = platform.machine()
        self._os_processor = platform.processor()


    @property
    def is_bsd(self) -> bool:
        return self._os_system in OsInfo._bsd_variants

    @property
    def is_linux(self) -> bool:
        return self._os_platform.lower().startswith("linux") or self._os_system.lower() == "linux"

    @property
    def is_macos(self) -> bool:
        return self._os_platform.lower() == "darwin" or self._os_system.lower() == "darwin"

    @property
    def is_windows(self) -> bool:
        return self._os_platform.lower() == "win32" or self._os_system.lower() == "windows"

    @property
    def is_wsl(self) -> bool:
        return self.is_linux() and self._os_release.lower().find('microsoft') >= 0

    @property
    def canonical_os_name(self) -> str:
        """Return the canonical name of the operating system.

        :rtype: str
        """
        if self.is_windows:
            return 'windows'
        if self.is_macos:
            return 'macos'
        if self.is_linux:
            if self.is_wsl:
                return 'wsl'
            else:
                return 'linux'
        if self.is_bsd:
            return self._os_name.lower()

        return 'unknown'


def remove_duplicates_from_list(in_list: list[Any]) -> list[Any]:
    """Remove duplicates from a list, preserving order.

        :param list[Any] in_list: The list to remove duplicates from.
        :return: A new list with the duplicates removed.
        :rtype: list[Any]
    """
    seen_entries = set()  # all entries previously seen
    return [x for x in in_list if not (x in seen_entries or seen_entries.add(x))]


def resolve_wildcards(filenames: list[str]) -> list[Path]:
    """Resolve unique Paths from a list containing wildcards.

        Item order of the input list is preserved in result.

        :param list[str] filenames: A list of filenames that may contain
             wildcards.
        :return: A list of unique Paths.
        :rtype: list[Path]

        .. versionchanged:: 1.5.2 Item order is preserved.
    """
    # Resolve wildcards.
    names = []
    for n in filenames:
        i = glob.glob(n)
        if i:
            names.extend(i)
    names = remove_duplicates_from_list(names)

    # Convert strings to paths.
    paths = [Path(n) for n in names]
    return paths


def rotate_file(file_name: str,
                file_dir: Path,
                rotation: int,
                compress_level: int = 9,
                fn_logger: logging.Logger = logging.getLogger(__name__)):
    """Rotate a (log-)file.

        The parameter rotation specifies the number of copies that will be
        retained.

        When rotating, file_name.rotation-1 is renamed to file_name.rotation
        for all values of rotation down to 1.

        The initial file (with no .rotation suffix) gains a suffix. If
        compress_level > 0, the initial file is also compressed.

        :param str file_name: Name of the log file.
        :param Path file_dir: Path to the directory containing the log file.
        :param int rotation: Number of copies to rotate. 0 to disable.
        :param int compress_level: Zlib compression level with 0 being no
            compression and 9 the highest possible.
        :param logging.Logger fn_logger: Logger used by the function.
    """
    fn_logger.debug('rotate_log(%s, %s, %d)', file_name, file_dir, rotation)
    if rotation == 0:
        fn_logger.debug('rotate_log(): nothing to rotate.')
        return

    max_rotation = rotation
    while rotation >= 0:
        if rotation == 0:
            candidate = file_dir / file_name
            if not candidate.exists():
                fn_logger.debug('rotate_log(): rotation 0 does not exists. Done.')
                return
            candidate_is_compressed = False
        else:
            # Find previously rotated candidate.
            candidate = file_dir / f'{file_name}.{rotation}'
            candidate_is_compressed = False
            if not candidate.exists():
                candidate = file_dir / f'{file_name}.{rotation}.gz'
                candidate_is_compressed = True
                if not candidate.exists():
                    fn_logger.debug('rotate_log(): rotation %d does not exist.', rotation)
                    rotation -= 1
                    continue

        # We have found a match.
        fn_logger.debug('rotate_log(): found candidate %s.', candidate)
        if rotation == 0:
            # All other rotation copies have been renamed, now the last logfile
            # must be compressed (optional) and renamed.
            compress_path = file_dir / file_name
            if compress_level > 0:
                new_name = f'{file_name}.1.gz'
                new_path = file_dir / new_name
                fn_logger.debug('rotate_log(): compress %s -> %s.', compress_path, new_path)
                with open(compress_path, 'rb') as f_in:
                    # Get the modification time of the file we are rotating
                    # so it will be stored correctly in the gzip archive.
                    statinfo = os.stat(compress_path)
                    with open(new_path, 'wb') as f_out:
                        with gzip.GzipFile(file_name,
                                           'wb',
                                           compress_level,
                                           f_out,
                                           mtime=statinfo.st_mtime) as f_zip:
                            shutil.copyfileobj(f_in, f_zip)

                fn_logger.debug('rotate_log(): unlink %s.', compress_path)
                compress_path.unlink()
            else:
                new_path = file_dir / f'{file_name}.{rotation + 1}'
                fn_logger.debug('rotate_log(): rename %s -> %s', candidate, new_path)
                candidate.rename(new_path)
        elif rotation == max_rotation:
            # We have found the maximum rotation file. Delete it to make room.
            fn_logger.debug('rotate_log(): unlinking old %s.', candidate)
            candidate.unlink()
        elif rotation > 0:
            # Rename existing rotation to next higher.
            if candidate_is_compressed:
                new_path = file_dir / f'{file_name}.{rotation + 1}.gz'
            else:
                new_path = file_dir / f'{file_name}.{rotation + 1}'
            fn_logger.debug('rotate_log(): rename %s -> %s', candidate, new_path)
            candidate.rename(new_path)
        rotation -= 1


def round_down(n: int, m: int) -> int:
    """Round the given number *n* down to the nearest multiple of *m*.

        :param int n: number to round
        :param int m: multiple to round to
        :return: n rounded down to a multiple of m.
        :rtype: int
    """
    return n & ~(m - 1)


def round_up(n: int, m: int) -> int:
    """Round the given number *n* up to the nearest multiple of *m*.

        :param int n: number to round
        :param int m: multiple to round to
        :return: n rounded up to a multiple of m.
        :rtype: int
    """
    return (n + m - 1) & ~(m - 1)


@contextlib.contextmanager
def std_open(filename: str | Path | None = None,
             mode: str = 'rt',
             encoding: str = 'utf-8',
             newline: str | None = None):
    """Open either a file or stdin/stdout for use with `with`.

        If the filename is None or '-' then stdin or stdout (depending on the
        mode) are used.
        Otherwise, the file is used. It is closed when the `with` block is done.

        If the filename ends with a well-known compression format extension,
        this compression format is used:

        .bz2
            bzip2 compressed file, uses :py:func:`bz2.open`
        .gz
            GZip compressed file, uses :py:func:`gzip.open`.
        .xz
            LZMA compressed file, uses :py:func:`lzma.open`.

        :param str|Path|None filename: A filename, Path-like, '-', or `None`.
        :param str mode: The mode to use for :func:`open`. Valid modes are:
            'rb', 'ab', 'wb', 'xb' for binary mode, or 'rt', 'at', 'wt', or 'xt'
            for text mode. The default is 'rt'.
        :param str encoding: The encoding to pass to :func:`open`. Defaults to 'utf-8'.
        :param str newline: Universal newline mode. Passed to :func:`open`.

        .. note::

            The modes 'a', 'r', 'w', and 'x' default to text mode for
            uncompressed, and to binary mode for compressed formats. I
            recommend avoiding them. Always express your programs intent
            explicitly to avoid problems.

        Example
        -------
        .. code-block:: python

            with std_open(fn, 'rt') as f:
                c = f.read()

        .. versionchanged:: 1.3.1
           Default mode changed to 'r' to minimize chances of accidental file clobbering.
        .. versionchanged:: 1.4.2
           Add parameter `newline`.
        .. versionchanged:: 1.5.1
           Add ability to read/write compressed files.
           Allow Path-like object for the filename.
    """
    format_map = {
        '.bz2': bz2.open,
        '.bzip2': bz2.open,
        '.gz': gzip.open,
        '.gzip': gzip.open,
        '.xz': lzma.open,
    }  # format_map
    open_func = open

    if not filename or filename == '-':
        if 'r' in mode:  # pylint: disable=else-if-used
            fh = sys.stdin
        else:  # 'w'
            fh = sys.stdout
    else:
        file_path = Path(filename)
        file_suffix = file_path.suffix
        if file_suffix in format_map:
            open_func = format_map[file_suffix]

        if 'b' in mode:
            fh = open_func(file_path, mode=mode)
        else:
            # Text mode allows more options than binary mode.
            fh = open_func(file_path,
                           mode=mode,
                           encoding=encoding,
                           newline=newline)

    try:
        yield fh
    finally:
        if fh is not sys.stdin and fh is not sys.stdout:
            fh.close()


def string_from_format(fmt: str, timestamp: datetime = None) -> str:
    """Given a strftime()-like format, expand into a string.

        Additional formats:
        - %P - name of the application

        :param str fmt: Format string to apply.
        :param datetime timestamp: Timestamp to format or `None` to use the current time.
        :return: Resulting string.
    """
    fmt2 = ''
    if '%P' in fmt:
        start = 0
        pos = fmt.find('%P', start)
        while pos >= 0:
            length = pos - start
            fmt2 += fmt[start:length] + Path(sys.argv[0]).stem
            start = pos + 2
            pos = fmt.find('%P', start)
        fmt2 += fmt[start:]
    else:
        fmt2 = fmt

    if timestamp is None:
        timestamp = datetime.now()
    return timestamp.strftime(fmt2)


def string_to_snake_case(string: str) -> str:
    """Convert a string to snake_case.

    :param str string: String to convert.
    :rtype str:
    :return: Converted snake_case string.

    .. seealso:: https://stackoverflow.com/a/1176023
    """
    # Edge case of an acronym followed by another word (e.g. "HTTPResponse"
    # -> "HTTP_Response") OR the more normal case of an initial lowercase word
    # followed by a capitalized word (e.g. "getResponse" -> "get_Response").
    string = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    # Replace double __ with single _ (e.g. __HTTP_Response -> _HTTP_Response).
    string = re.sub('__([A-Z])', r'_\1', string)
    # Normal case of two non-acronyms (e.g. "ResponseCode" -> "Response_Code")
    string = re.sub('([a-z0-9])([A-Z])', r'\1_\2', string)
    # Convert everything to lower case.
    return string.lower()


def true_stem(path: Path) -> str:
    """Return the true stem (e.g. the name without suffixes) of the Path.

        :param Path path: A Path object.
        :rtype: str
        :return: The final file or directory name without any suffixes.

        Example
        -------

        .. code-block:: python

            print(true_stem(Path('/var/tmp/abc.tar.gz')))
            abc

    """
    ts = path.stem
    while path.suffixes:
        ts = path.stem
        path = Path(ts)
    return ts


def wrapping_counter_delta(first: int, second: int, max_value: int) -> int:
    """Calculate the difference between two values of a wrapping counter.

    :param int first: The first counter value.
    :param int second: The second (later) counter value.
    :param int max_value: The maximum value the counter can reach plus one.
        For fixed-length integers, this is 2**<num_bits>.
    :returns: The difference between the two counter values.

    Example for a 4 bit counter (range 0..15)::

         0             0            0             0 1             1
         0             f            0             f 0             f
        |---------------|          |---------------|---------------|
           ^        ^       ===>               ^      ^
           2nd      1st                        1st    2nd

        first  = 0x0c (= 12)
        second = 0x03 (=  3)
        delta = 0x03 + 0x10 - 0x0c = 0x07
                   3 +   16 -   12 =    7

    .. versionadded:: 1.5.0
    """
    if first <= second:
        delta = second - first
    else:
        delta = second + max_value - first
    return delta


if __name__ == '__main__':
    if sys.version_info < (3, 9):
        sys.stderr.write('FATAL ERROR: Python 3.9 or later is required.\n')
        sys.exit(1)



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-deadlinequeue-py
"""A queue of items sorted by the time they are due.
"""

import heapq
import logging
import sys
import threading
import time
from typing import Any


class DeadlineQueue:
    """A queue of items sorted by the time they are due.

    An item is due when the deadline expires.

    To add an element that will expire immediately, use timedue==0.

    The queue consists of tuples of the form (deadline, idx, item). The
    lowest deadline is sorted to the front of the queue.

    The deadline is based on time.perf_counter().

    The role of the idx variable is to properly order items with the same
    priority level. By keeping a constantly increasing idx, the items
    will be sorted according to the order in which they were inserted.
    """

    def __init__(self, logger: logging.Logger = logging.getLogger(__name__)):
        """Initialize the instance.

        :param logging.Logger logger: The logger used by the instance.
        """
        # A counter used to create unique index numbers used for sorting."""
        self._count: int = 0
        # A condition used to lock access to the queue.
        self._cv = threading.Condition()
        # The logger used to output information.
        self._logger: logging.Logger = logger
        # A list of items.
        self._queue: list[tuple[float, int, Any]] = []

    def __repr__(self) -> str:
        """Return a human-readable string representation of the instance."""
        s = '['
        not_first = False
        with self._cv:
            for item in self._queue:
                if not_first:
                    s = s + f', {item[2]} @ {item[0]}'
                else:
                    not_first = True
                    s = s + f'{item[2]} @ {item[0]}'
        s = s + ']'
        return s

    def put(self, item_tuple: tuple[float, Any]):
        """Place a tuple in the queue.

        The item tuple is (deadline, item). A lower deadline is sorted earlier
        in the queue.

        :param item_tuple tuple[float, Any]: An item to place in the queue with
            a deadline in seconds.
        """
        deadline = item_tuple[0]
        item = item_tuple[1]
        # Lock access to the queue by grabbing the condition.
        with self._cv:
            heapq.heappush(self._queue, (deadline, self._count, item))
            self._logger.debug(
                'put(#%u - %s @ %.6f s at %.6f s',
                self._count, item, deadline, time.perf_counter())
            self._count += 1
            # Notify anyone waiting on the condition that the queue has
            # changed.
            self._cv.notify()

    def get(self) -> tuple[float, int, Any]:
        """Return the first element in the queue and remove it.

        The element will be tuple (deadline, count, item).

        :return: First element in the queue.
        :rtype  tuple[float, int, Any]:
        """
        with self._cv:
            # Wait until there is at least one element in the queue.
            while len(self._queue) == 0:
                self._cv.wait()
            element = heapq.heappop(self._queue)
        self._logger.debug(
            'get(#%u - %s @ %.6f s at %.6f s',
            element[1], element[2], element[0], time.perf_counter())
        return element

    def peek(self) -> tuple[float, int, Any]:
        """Returns the first element in the queue without removing it.

        The element will be tuple (deadline, count, item).

        :return: First element in the queue.
        :rtype  tuple[float, int, Any]:
        """
        with self._cv:
            # Wait until there is at least one element in the queue.
            while len(self._queue) == 0:
                self._cv.wait()
            element = self._queue[0]
        self._logger.debug(
            'peek(#%u - %s @ %.6f s',
            element[1], element[2], element[0])
        return element

    def qsize(self) -> int:
        """Return the number of elements in the queue.

            :return: Number of elements in the queue.
            :rtype int:
        """
        with self._cv:
            return len(self._queue)


if __name__ == '__main__':
    from threading import Barrier, BrokenBarrierError, Event, Thread

    start_barrier = Barrier(2, timeout=1)   # Start running producer and consumer.
    end_barrier = Barrier(3, timeout=60)    # Wait to terminate.
    terminate_event = Event()

    # Object that signals shutdown
    _sentinel = object()

    # The time offsets (in microseconds) when an event will be fired.
    event_times = [0, 0.5, 0.2, 0.1, 0, 0.01, 0.02, 0, 0.099]

    def producer(dq_out):
        """Place a new data item into the queue every interval seconds for number items."""
        start_barrier.wait()
        index = 0
        start_micros = time.perf_counter() + 0.010
        for et in event_times:
            if et != 0:
                et = start_micros + et
            # Produce some data
            data = f'Some value {index}'
            dq_out.put((et, data))
            print(f'Put: {data} @ {et:.6f} s')
            index = index + 1
        dq_out.put((0, 'abc'))
        print('Put: sentinel')
        dq_out.put((sys.float_info.max, _sentinel))
        print(f'Queue: {dq_out}')
        print(f'Producer is done, created {index} elements (plus sentinel).')
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            print('producer() end_barrier broken')

    def consumer(dq_in):
        """Get data items from the queue and process them."""
        start_barrier.wait()
        items = 0
        while True:
            # Peek at the data.
            (deadline, idx, data) = dq_in.peek()
            current_s = time.perf_counter()
            if deadline > current_s:
                if terminate_event.is_set():
                    break
                continue

            # Get the first element in the queue.
            (deadline, idx, data) = dq_in.get()
            items = items + 1

            if data is _sentinel:
                print('Get: sentinel')
                break
            print(f'Get: {data} @ {deadline:.6f} s (qsize {dq_in.qsize()})'
                  f' delta_s {current_s - deadline:.6f}')

        # Indicate completion
        print(f'Consumer is done, qsize {dq_in.qsize()}.')
        print(f'Queue: {dq_in}')
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            print('consumer() end_barrier broken')

    def demo():
        """Demonstration of the DeadlineQueue and how to use it."""
        # Create the shared queue and launch both threads
        q = DeadlineQueue()
        t1 = Thread(target=consumer, args=(q,))
        t2 = Thread(target=producer, args=(q,))
        t1.start()
        t2.start()
        # Wait for all produced items to be consumed
        try:
            end_barrier.wait(timeout=5)
        except BrokenBarrierError:
            print('demo() end_barrier broken')
            terminate_event.set()
        print('Both threads are done. Terminating')
        sys.exit(0)

    # Perform demonstration.
    demo()



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-utilities-convert-py

from datetime import datetime, timezone
import math

import pytest
from pymisclib.utilities import convert_dotnet_datetime_ticks_to_datetime, convert_ms_since_epoch_to_datetime


dotnet_test_vectors = [
    # DateTime.MaxValue
    (0, datetime.fromisoformat('0001-01-01T00:00:00Z')),
    # Int32.MaxValue should be .7483647 but is 748365 due to rounding errors
    (2147483647, datetime.fromisoformat('0001-01-01T00:03:34.748365Z')),
    # 1 hour
    (36000000000, datetime.fromisoformat('0001-01-01T01:00:00Z')),
    # Unix epoch
    (621355968000000000, datetime.fromisoformat('1970-01-01T00:00:00Z')),
    (630822816000000000, datetime.fromisoformat('2000-01-01T00:00:00Z')),
    (633979008000000000, datetime.fromisoformat('2010-01-01T00:00:00Z')),
    (637134336000000000, datetime.fromisoformat('2020-01-01T00:00:00Z')),
    (638712864000000000, datetime.fromisoformat('2025-01-01T00:00:00Z')),
    # 1 day ago
    (638786877890000000, datetime.fromisoformat('2025-03-27T15:56:29.000Z')),
    # 1 hour ago
    (638787705890000000, datetime.fromisoformat('2025-03-28T14:56:29.000Z')),
    # now
    (638787741890000000, datetime.fromisoformat('2025-03-28T15:56:29.000Z')),
    # datime.MAXYEAR
    (3155378975999990000, datetime.fromisoformat('9999-12-31T23:59:59.999Z'))
] # dotnet_test_vectors


@pytest.mark.parametrize("ticks,expected", dotnet_test_vectors)
def test_convert_dotnet_datetime_ok(ticks, expected):
    dt = convert_dotnet_datetime_ticks_to_datetime(ticks, timezone.utc)
    # We could naively compare dt to expected, but because of
    # rounding errors, this can fail for the milliseconds.
    assert dt.year == expected.year
    assert dt.month == expected.month
    assert dt.day == expected.day
    assert dt.hour == expected.hour
    assert dt.minute == expected.minute
    assert dt.second == expected.second
    assert math.ceil(dt.microsecond * 1000.0) == math.ceil(dt.microsecond * 1000.0)


def test_convert_dotnet_datetime_maxvalue():
    # DateTime.MaxValue
    with pytest.raises(OverflowError):
        dt = convert_dotnet_datetime_ticks_to_datetime(
            3155378975999999999, timezone.utc)


unix_test_vectors = [
    # Epoch
    (0, datetime.fromisoformat('1970-01-01T00:00:00Z')),
    # 0xffffffff
    (4294967295, datetime.fromisoformat('1970-02-19T17:02:47.295Z')),
    # 0x100000000
    (4294967296, datetime.fromisoformat('1970-02-19T17:02:47.296Z')),
    # Date within digits.
    (119731017000, datetime.fromisoformat('1973-10-17T18:36:57Z')),
    # Billenium
    (1000000000000, datetime.fromisoformat('2001-09-09T01:46:40Z')),
    # 1234567890 seconds
    (1234567890123, datetime.fromisoformat('2009-02-13T23:31:30.123Z')),
    # PI
    (3141592653589, datetime.fromisoformat('2069-07-21T00:37:33.589Z')),
    # datetime.MAXYEAR
    (253402300799999, datetime.fromisoformat('9999-12-31T23:59:59.999Z')),
]  # unix_test_vectors


@pytest.mark.parametrize("ticks,expected", unix_test_vectors)
def test_convert_epoch_ms_ok(ticks, expected):
    dt = convert_ms_since_epoch_to_datetime(ticks)
    # The naive comparison should work.
    assert dt == expected



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-enumaction-py
import argparse
import enum

from pymisclib.enumaction import EnumAction
import pytest


def name_helper(enum_class: enum.Enum,
                argv: list[str|int],
                expected: argparse.Namespace | None):
    """Test EnumAction using the name."""
    parser = argparse.ArgumentParser(exit_on_error=False)
    parser.add_argument('--name', '-n', type=enum_class,
                        use='name', action=EnumAction)
    if expected is None:
        with pytest.raises(argparse.ArgumentError):
            args = parser.parse_args(argv)
    else:
        args = parser.parse_args(argv)
        assert args == expected


def value_helper(enum_class: enum.Enum,
                argv: list[str|int],
                expected: argparse.Namespace | None):
    """Test EnumAction using the value."""
    parser = argparse.ArgumentParser(exit_on_error=False)
    parser.add_argument('--value', '-v', type=enum_class,
                        use='value', action=EnumAction)
    if expected is None:
        with pytest.raises(argparse.ArgumentError):
            args = parser.parse_args(argv)
    else:
        args = parser.parse_args(argv)
        assert args == expected


class TestEnum(enum.Enum):
    E1 = 'e1'
    E2 = 'e2'


def test_enumaction_name_enum_ok():
    """Test good cases of using the name of an enum.Enum subclass."""
    name_helper(TestEnum,
                ['--name', 'E1'],
                argparse.Namespace(name=TestEnum.E1))
    name_helper(TestEnum,
                ['-n', 'E2'],
                argparse.Namespace(name=TestEnum.E2))


def test_enumaction_name_enum_err():
    """Test error cases of using the name of an enum.Enum subclass."""
    name_helper(TestEnum, ['--name', 'e1'], None)
    name_helper(TestEnum, ['-n', 'e1'], None)
    name_helper(TestEnum, ['-n', 'does_not_exist'], None)


def test_enumaction_value_enum_ok():
    """Test good cases of using the value of an enum.Enum subclass."""
    value_helper(TestEnum,
                ['--value', 'e1'],
                argparse.Namespace(value=TestEnum.E1))
    value_helper(TestEnum,
                ['-v', 'e2'],
                argparse.Namespace(value=TestEnum.E2))


def test_enumaction_value_enum_err():
    """Test error cases of using the value of an enum.Enum subclass."""
    value_helper(TestEnum, ['--value', 'E1'], None)
    value_helper(TestEnum, ['-v', 'E1'], None)
    value_helper(TestEnum, ['-v', 'does_not_exist'], None)


class TestIntEnum(enum.IntEnum):
    I1 = 1
    I2 = 2


def test_enumaction_name_int_enum_ok():
    """Test good cases of using the name of an enum.IntEnum subclass."""
    name_helper(TestIntEnum,
                ['--name', 'I1'],
                argparse.Namespace(name=TestIntEnum.I1))
    name_helper(TestIntEnum,
                ['-n', 'I2'],
                argparse.Namespace(name=TestIntEnum.I2))


def test_enumaction_name_int_enum_err():
    """Test error cases of using the name of an enum.IntEnum subclass."""
    name_helper(TestEnum, ['--name', 'i1'], None)
    name_helper(TestEnum, ['-n', '1'], None)
    name_helper(TestEnum, ['-n', 'does_not_exist'], None)


def test_enumaction_value_int_enum_err():
    """All cases of using the value of an enum.IntEnum subclass are error cases."""
    value_helper(TestIntEnum, ['--value', '1'], None)
    value_helper(TestIntEnum, ['--value', 'does_not_exist'], None)

    with pytest.raises(TypeError):
        # Python argparse can not handle non-string arguments.
        value_helper(TestIntEnum, ['-v', 1], None)


class TestFlagEnum(enum.Flag):
    F1 = 1
    F2 = 2


def test_enumaction_name_flag_ok():
    """Test good cases of using the name of an enum.IntEnum subclass."""
    name_helper(TestFlagEnum,
                ['--name', 'F1'],
                argparse.Namespace(name=TestFlagEnum.F1))
    name_helper(TestFlagEnum,
                ['-n', 'F2'],
                argparse.Namespace(name=TestFlagEnum.F2))


def test_enumaction_name_flag_err():
    """Test error cases of using the name of an enum.Flag subclass."""
    name_helper(TestEnum, ['--name', 'f1'], None)
    name_helper(TestEnum, ['-n', '1'], None)
    name_helper(TestEnum, ['-n', 'does_not_exist'], None)


def test_enumaction_value_flag_err():
    """All cases of using the value of an enum.FlagEnum subclass are error cases."""
    value_helper(TestIntEnum, ['--value', '1'], None)
    value_helper(TestIntEnum, ['--value', 'does_not_exist'], None)

    with pytest.raises(TypeError):
        # Python argparse can not handle non-string arguments.
        value_helper(TestIntEnum, ['-v', 1], None)


class TestIntFlagEnum(enum.IntFlag):
    IF1 = 1
    IF2 = 2
    IF4 = 4


def test_enumaction_name_int_flag_ok():
    """Test good cases of using the name of an enum.IntFlag subclass."""
    name_helper(TestIntFlagEnum, ['--name', 'IF1'],
                argparse.Namespace(name=TestIntFlagEnum.IF1))
    name_helper(TestIntFlagEnum, ['-n', 'IF2'],
                argparse.Namespace(name=TestIntFlagEnum.IF2))

def test_enumaction_name_int_flag_err():
    """Test error cases of using the name of an enum.IntFlag subclass."""
    name_helper(TestIntEnum, ['--name', '1'], None)
    name_helper(TestIntEnum, ['--name', 'does_not_exist'], None)
    # Multiple flags are not supported.
    name_helper(TestIntFlagEnum, ['-n', 'IF1|IF4'], None)

    with pytest.raises(TypeError):
        # Python argparse can not handle non-string arguments.
        value_helper(TestIntEnum, ['-v', 3], None)



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024, 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-chunkystreamsocket-py
"""Unit tests for ChunkyStreamSocket."""

from dataclasses import dataclass
import socket
import threading
import time

import pytest
from pymisclib.chunkystreamsocket import ChunkyStreamSocket


@dataclass
class EchoServer:
    """Simple test server that accepts connections and echos all input."""
    _backlog: int = 0  # allow no connection backlog
    _loop_time: float = 0.1
    _running: bool = False
    _sock: socket.socket|None = None
    _thread: threading.Thread|None = None

    @property
    def port(self) -> int|None:
        """Return the server port."""
        if self._sock is None:
            return None
        return self._sock.getsockname()[1]

    @property
    def running(self) -> bool:
        """True if the loop is running."""
        return self._running

    @property
    def sock(self) -> socket.socket|None:
        """Return the OS socket or None if there is none."""
        return self._sock

    def _client_loop(self,
                     client_sock: socket.socket):
        """Handle client connections while running."""
        client_sock.settimeout(self._loop_time)
        while self._running:
            try:
                msg = client_sock.recv(1)
                client_sock.sendall(msg)
            except socket.timeout:
                pass
            except OSError:
                break
        try:
            client_sock.close()
        except OSError:
            pass

    def _run_loop(self):
        """While running, accept client connections."""
        if self._running:
            raise RuntimeError('Server is already running.')
        self._sock = socket.socket()
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind(('localhost', 0))  # bind to a free port
        self.sock.listen(self._backlog)
        self._running = True
        self._sock.settimeout(0.1)  # all socket operations time out after 100 ms
        while self._running:
            try:
                client_sock, client_addr = self._sock.accept()
                self._client_loop(client_sock)
            except socket.timeout:
                pass  # ignore to allow checking for shutdown
        self._sock.close()
        self._sock = None

    def run(self):
        """Run server in a new thread."""
        self._thread = threading.Thread(target=self._run_loop)
        self._thread.start()
        while not self._running:
            time.sleep(0.001)

    def stop(self):
        """Stop the server."""
        self._running = False
        self._thread.join(5.0)
        self._thread = None


@pytest.fixture(autouse=True)
def server() -> EchoServer:
    """Fixture to start an EchoServer before running the test function."""
    server = EchoServer()
    server.run()
    yield server  # run test function
    server.stop()


@pytest.fixture()
def used_port() -> int:
    """Fixture that creates a socket to block a port.

    The fixture returns the port number. Since the socket on the port
    is not serviced, connections to it will be refused.
    """
    s = socket.socket()
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(('localhost', 0))  # bind to a free port
    yield s.getsockname()[1]  # run test function
    s.close()


def connect_and_check(es: EchoServer,
                      css: ChunkyStreamSocket,
                      timeout: float|None):
    """Connect and validate the connection.

    :param EchoServer es: EchoServer instance to connect to.
    :param ChunkyStreamSocket css: Socket instance to connect from.
    :param float timeout: Timeout to wait for the connection.
    """
    css.connect(timeout)
    addr, port = css.sock.getpeername()
    assert addr == '127.0.0.1'
    assert port == es.port


@pytest.mark.parametrize('timeout', [1.0, 0, None])
def test_connect_ok(server, timeout):
    """Test connection to an existing server for various client socket timeouts."""
    with ChunkyStreamSocket(host='127.0.0.1', port=server.port) as c:
        try:
            connect_and_check(server, c, timeout=timeout)
        except BlockingIOError:
            # A non-blocking connection attempt, e.g. connect(timeout=0),
            # will almost always fail on the first try. So we give it some
            # time and try again.
            assert timeout == 0
            time.sleep(0.1)  # since this is local, 100 ms should be plenty.
            connect_and_check(server, c, timeout=timeout)


@pytest.mark.parametrize('timeout', [0.2, 0, None])
def test_connect_err_connection(server, timeout):
    """Test connection with a timeout to a non-existing IP address."""
    with ChunkyStreamSocket(host='192.0.2.1', port=server.port) as c:
        with pytest.raises(ConnectionError):
            try:
                c.connect(timeout)
            except BlockingIOError:
                # A non-blocking connection attempt, e.g. connect(timeout=0),
                # will almost always fail on the first try. So we give it some
                # time and try again.
                assert timeout == 0
                time.sleep(0.1)  # since this is local, 100 ms should be plenty.
                c.connect(timeout)


@pytest.mark.parametrize('timeout', [0.2, 0, None])
def test_connect_err_refused(server, timeout, used_port):
    """Test connection that is refused."""
    with ChunkyStreamSocket(host='127.0.0.1', port=used_port) as c:
        with pytest.raises(ConnectionError):
            try:
                c.connect(timeout)
            except BlockingIOError:
                # A non-blocking connection attempt, e.g. connect(timeout=0),
                # will almost always fail on the first try. So we give it some
                # time and try again.
                assert timeout == 0
                time.sleep(0.1)  # since this is local, 100 ms should be plenty.
                c.connect(timeout)


def test_connect_err_timout(server):
    """Test connection request timing out."""
    # Implementation note: this relies on socket.listen(backlog=N)
    # to actually limit the socket backlog to N. This assumption
    # is not necessarily true, so we set the backlog to zero and
    # create a large number of connections.
    css = []
    with pytest.raises(TimeoutError):
        for i in range(4096):
            c = ChunkyStreamSocket(host='127.0.0.1', port=server.port)
            c.connect(timeout=0.1)
            css.append(c)


def test_method_connection_ok(server):
    """Test the connection() method."""
    with ChunkyStreamSocket(host='127.0.0.1', port=server.port) as c:
        assert isinstance(c.connection, str)
        c.connect()
        assert isinstance(c.connection, str)
        c.close()
        assert isinstance(c.connection, str)


def test_property_connected_ok(server):
    """Test the connected property."""
    with ChunkyStreamSocket(host='127.0.0.1', port=server.port) as c:
        assert not c.connected
        c.connect()
        assert c.connected
        c.close()
        assert not c.connected


def test_property_server_socket_ok():
    with ChunkyStreamSocket(host='127.0.0.1', port=0) as c:
        assert not c.server_socket
        c.bind_and_listen(0)
        assert c.server_socket
        c.close()
        assert not c.server_socket


@pytest.mark.parametrize('timeout', [0.1, 0, None])
def test_send_and_receive_ok(server, timeout):
    """Test sending and receiving messages."""
    messages = [
        b'',
        b'abc',
        b'\x00\x01\x02\x03\x04',
        bytes(range(256)),
        b'The quick brown fox jumps over the lazy dog' * 1024,
    ]
    with ChunkyStreamSocket(host='127.0.0.1', port=server.port) as c:
        try:
            connect_and_check(server, c, timeout=timeout)
        except BlockingIOError:
            # A non-blocking connection attempt, e.g. connect(timeout=0),
            # will almost always fail on the first try. So we give it some
            # time and try again.
            assert timeout == 0
            time.sleep(0.1)  # since this is local, 100 ms should be plenty.
            connect_and_check(server, c, timeout=timeout)
        for message in messages:
            c.send(message)
            rx_msg = c.recv(len(message))
            assert rx_msg == message


def test_method_close_ok():
    """Test the close() method."""
    c = ChunkyStreamSocket(host='127.0.0.1', port=0)
    c.close()
    assert c.sock is None



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges
# SPDXID: SPDXRef-test-ringbuffer-py
import contextlib
import io
import os
import sys
from pathlib import Path
import random
import tempfile
import unittest

from pymisclib.utilities import std_open


# Create test cases dynamically.
# See https://stackoverflow.com/a/2799009 for how this works.
test_compressed = [
    'file.bz2',
    'file.gz',
    'file.xz',
]
test_uncompressed = [
    'file',  # no suffix
    'file.md',  # markdown suffix
    'file.txt',  # text suffix
    'file.unknown',  # unknown suffix
]

FILE_SIZE = 1024

def generate_random_unicode_string(length: int):
    """Generate a string filled with random unicode code points.

    :param int length: Length of the string to generate.

    :see: https://stackoverflow.com/a/21666621
    """
    # Update this to include code point ranges to be sampled
    include_ranges = [
        ( 0x0021, 0x0021 ),
        ( 0x0023, 0x0026 ),
        ( 0x0028, 0x007E ),
        ( 0x00A1, 0x00AC ),
        ( 0x00AE, 0x00FF ),
        ( 0x0100, 0x017F ),
        ( 0x0180, 0x024F ),
        ( 0x2C60, 0x2C7F ),
        ( 0x16A0, 0x16F0 ),
        ( 0x0370, 0x0377 ),
        ( 0x037A, 0x037E ),
        ( 0x0384, 0x038A ),
        ( 0x038C, 0x038C ),
    ]

    alphabet = [
        chr(code_point) for current_range in include_ranges
            for code_point in range(current_range[0], current_range[1] + 1)
    ]
    return ''.join(random.choice(alphabet) for i in range(length))


class FileTest(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        if self.tmp_dir:
            self.tmp_dir.cleanup()
            self.tmp_dir = None


class TestStdOpenCompressed(FileTest):
    pass


class TestStdOpenUncompressed(FileTest):
    pass


class TestStdOpenStdinStdout(FileTest):

    def test_stdin_text(self):
        """Read a text from stdin."""
        text = generate_random_unicode_string(FILE_SIZE)
        old_stdin = sys.stdin
        sys.stdin = io.StringIO(text)
        with std_open('-', 'rt') as f:
            read_text = f.read(FILE_SIZE)
        sys.stdin = old_stdin
        self.assertEqual(text, read_text)

    def test_stdout_text(self):
        text = generate_random_unicode_string(FILE_SIZE)
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            with std_open('-', 'wt') as f:
                f.write(text)
        self.assertEqual(text, stdout.getvalue())


def create_tests_compressed_binary(filename: str):
    """Return a test function for the given test vector (e.g. filename) to
    test compressed binary IO.

    :param str filename: Name of the file to work on. Will be located in a
        temporary directory.
    """
    def test_compressed_binary(self):
        """Test writing to/writing from a compressed file in binary mode.
        Make sure that reading uncompressed fails.
        """
        file_path = Path(self.tmp_dir.name) / filename
        binary = os.urandom(FILE_SIZE)
        with std_open(file_path, 'wb') as f:
            f.write(binary)
        with std_open(file_path, 'rb') as f:
            read_binary = f.read(FILE_SIZE)
            self.assertEqual(read_binary, binary)
        with open(file_path, 'rb') as f:
            read_binary_uncompressed = f.read(FILE_SIZE)
            self.assertNotEqual(read_binary_uncompressed, binary)

    return test_compressed_binary


def create_tests_compressed_text(filename: str):
    """Return a test function for the given test vector (e.g. filename) to
    test compressed text IO.

    :param str filename: Name of the file to work on. Will be located in a
        temporary directory.
    """
    def test_compressed_text(self):
        """Test writing to/writing from a compressed file in text mode.
        Make sure that reading uncompressed fails.
        """
        file_path = Path(self.tmp_dir.name) / filename
        text = generate_random_unicode_string(FILE_SIZE)
        with std_open(file_path, 'wt') as f:
            f.write(text)
        with std_open(file_path, 'rt') as f:
            read_text = f.read(FILE_SIZE)
            self.assertEqual(read_text, text)
        with open(file_path, 'rt') as f:
            try:
                read_text_uncompressed = f.read(FILE_SIZE)
            except UnicodeDecodeError:
                # Reading random binary data as Unicode text is likely to cause
                # an exception, so we'll take that as an indication that the
                # file really is compressed.
                read_text_uncompressed = ''
            self.assertNotEqual(read_text_uncompressed, text)

    return test_compressed_text


def create_tests_uncompressed_binary(filename: str):
    """Return a test function for the given test vector (e.g. filename) to
    test uncompressed binary IO.

    :param str filename: Name of the file to work on. Will be located in a
        temporary directory.
    """
    def test_uncompressed_binary(self):
        """Test writing to/writing from an uncompressed file in binary mode.
        Make sure that reading uncompressed (with plain open()) works.
        """
        file_path = Path(self.tmp_dir.name) / filename
        binary = os.urandom(FILE_SIZE)
        with std_open(file_path, 'wb') as f:
            f.write(binary)
        with std_open(file_path, 'rb') as f:
            read_binary = f.read(FILE_SIZE)
            self.assertEqual(read_binary, binary)
        with open(file_path, 'rb') as f:
            read_binary_uncompressed = f.read(FILE_SIZE)
            self.assertEqual(read_binary_uncompressed, binary)

    return test_uncompressed_binary


def create_tests_uncompressed_text(filename: str):
    """Return a test function for the given test vector (e.g. filename) to
    test uncompressed text IO.

    :param str filename: Name of the file to work on. Will be located in a
        temporary directory.
    """
    def test_uncompressed_text(self):
        """Test writing to a compressed file and then reading from it.
        Make sure that reading uncompressed fails.
        """
        file_path = Path(self.tmp_dir.name) / filename
        text = generate_random_unicode_string(FILE_SIZE)
        with std_open(file_path, 'wt') as f:
            f.write(text)
        with std_open(file_path, 'rt') as f:
            read_text = f.read(FILE_SIZE)
            self.assertEqual(read_text, text)
        with open(file_path, 'rt') as f:
            read_text_uncompressed = f.read(FILE_SIZE)
            self.assertEqual(read_text_uncompressed, text)

    return test_uncompressed_text


for file in test_compressed:
    # Create compressed binary test.
    tst_method = create_tests_compressed_binary(file)
    suffix = Path(file).suffix[1:]
    tst_method.__name__ = (f'test_compressed_binary_{suffix}')
    setattr(TestStdOpenCompressed, tst_method.__name__, tst_method)
    # Create compressed text test.
    tst_method = create_tests_compressed_text(file)
    suffix = Path(file).suffix[1:]
    tst_method.__name__ = (f'test_compressed_text_{suffix}')
    setattr(TestStdOpenCompressed, tst_method.__name__, tst_method)


for file in test_uncompressed:
    # Create uncompressed binary test.
    tst_method = create_tests_uncompressed_binary(file)
    suffix = Path(file).suffix[1:]
    tst_method.__name__ = (f'test_uncompressed_binary_{suffix}')
    setattr(TestStdOpenUncompressed, tst_method.__name__, tst_method)
    # Create uncompressed text test.
    tst_method = create_tests_uncompressed_text(file)
    suffix = Path(file).suffix[1:]
    tst_method.__name__ = (f'test_uncompressed_text_{suffix}')
    setattr(TestStdOpenUncompressed, tst_method.__name__, tst_method)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges
# SPDXID: SPDXRef-test-ringbuffer-py

import unittest

from pymisclib.littlejonny import LineStyle, Table, TableCell


class TestTableDimensions(unittest.TestCase):
    def test_empty_table_dimensions(self):
        table = Table()
        self.assertEqual(0, table.width)
        self.assertEqual(0, table.height)


class TestTableResize(unittest.TestCase):

    def setUp(self):
        self.table = Table(cells=[
            [TableCell('0;0'), TableCell('0;1'), TableCell('0;2'),],
            [TableCell('1;0'), TableCell('1;1'), TableCell('1;2'),],
            [TableCell('2;0'), TableCell('2;1'), TableCell('2;2'),],
        ])

    def table_integrity_check(self, table: Table):
        """Check the content of a table."""
        for r in range(table.height):
            row = table.cells[r]
            for c in range(table.width):
                cell = row[c]
                expect = f'{r};{c}'
                self.assertTrue(cell.content == expect or cell.content == '')

    def test_same_size(self):
        table = self.table
        table.resize(3, 3)
        self.assertEqual(table.height, 3)
        self.assertEqual(table.width, 3)
        self.table_integrity_check(table)

    def test_reduce_to_zero(self):
        table = self.table
        table.resize(0, 0)
        self.assertEqual(table.height, 0)
        self.assertEqual(table.width, 0)
        self.table_integrity_check(table)

    def test_increase_from_zero(self):
        """Test increasing a table with no size."""
        table = Table()
        table.resize(11, 7)
        self.assertEqual(7, table.height)
        self.assertEqual(11, table.width)

    def test_reduce_height(self):
        table = self.table
        table.resize(3, 2)
        self.assertEqual(table.height, 2)
        self.assertEqual(table.width, 3)
        self.table_integrity_check(table)

    def test_reduce_width(self):
        table = self.table
        table.resize(1, 3)
        self.assertEqual(table.height, 3)
        self.assertEqual(table.width, 1)
        self.table_integrity_check(table)

    def test_reduce_width_and_height(self):
        table = self.table
        table.resize(2, 1)
        self.assertEqual(table.height, 1)
        self.assertEqual(table.width, 2)
        self.table_integrity_check(table)

    def test_table_increase_height(self):
        table = self.table
        table.resize(3, 7)
        self.assertEqual(7, table.height)
        self.assertEqual(3, table.width)
        self.table_integrity_check(table)

    def test_cell_at(self):
        table = self.table
        for row in range(table.height):
            for col in range(table.width):
                cell = table.cell_at(col, row)
                self.assertEqual(f'{row};{col}', cell.content)

    def test_cell_at_error(self):
        table = self.table
        with self.assertRaises(IndexError):
            table.cell_at(0, -1)
        with self.assertRaises(IndexError):
            table.cell_at(-1, 0)
        with self.assertRaises(IndexError):
            table.cell_at(0, table.height)
        with self.assertRaises(IndexError):
            table.cell_at(table.width, 0)

    def test_column(self):
        table = self.table
        for col_idx in range(table.width):
            col = table.column(col_idx)
            self.assertEqual(table.height, len(col))
            row_idx = 0
            for cell in col:
                self.assertEqual(f'{row_idx};{col_idx}', cell.content)
                row_idx += 1

    def test_column_error(self):
        """Test invalid arguments to Table.column()."""
        table = self.table
        with self.assertRaises(IndexError):
            table.column(-1)
        with self.assertRaises(IndexError):
            table.column(table.height)

    def test_row(self):
        table = self.table
        for row_idx in range(table.height):
            row = table.row(row_idx)
            self.assertEqual(table.width, len(row))
            col_idx = 0
            for cell in row:
                self.assertEqual(f'{row_idx};{col_idx}', cell.content)
                col_idx += 1

    def test_row_error(self):
        """Test invalid arguments to Table.row()."""
        table = self.table
        with self.assertRaises(IndexError):
            table.row(-1)
        with self.assertRaises(IndexError):
            table.row(table.height)

    def test_resize_error(self):
        table = self.table
        with self.assertRaises(ValueError):
            table.resize(table.width, 0)
        self.assertEqual(table.height, 3)
        self.assertEqual(table.width, 3)
        with self.assertRaises(ValueError):
            table.resize(0, table.height)
        self.assertEqual(table.height, 3)
        self.assertEqual(table.width, 3)
        with self.assertRaises(ValueError):
            table.resize(table.width, -1)
        self.assertEqual(table.height, 3)
        self.assertEqual(table.width, 3)
        with self.assertRaises(ValueError):
            table.resize(-1, table.height)
        self.assertEqual(table.height, 3)
        self.assertEqual(table.width, 3)

#    def test_draw(self):
#        table = self.table
#        lines = table.draw(LineStyle.light)
#        self.assertEqual(1 + table.height * 2, len(lines))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-commandmenu-py
"""An interactive text console menu.

    # Example
    (pending)
"""

import logging
import readline


class CommandItem:
    """An item for a command menu."""
    def __init__(self, name: str,
                 default: bool = False):
        self.name = name
        self.default = default

    def __str__(self):
        """A human-readable description of the instance."""
        s = f'CommandItem<"{self.name}"'
        if self.default:
            s += ', default>'
        else:
            s += '>'
        return s


class CommandMenu:
    """Interactive text console menu."""
    # https://pymotw.com/2/readline/
    def __init__(self,
                 vocabulary: dict[str, list[str]] = {'quit': []},
                 prompt: str = '> ',
                 terminal_word: str = 'quit'):
        """Initialize the instance.

           :param dict[str, list[str]] vocabulary: A dictionary of primary terms
              with a list of secondary terms specific to the primary term.
           :param str prompt: The prompt displayed to tell the user that input
              is expected.
           :param str terminal_word: The primary term that will terminate the menu.
        """

        # Configuration members.
        self.prompt = prompt
        self.vocabulary = vocabulary
        self.terminal_word = terminal_word

        # Dynamic members.
        self.current_candidates = []

    def __str__(self):
        """A human-readable description of the instance."""
        s = 'CommandMenu<'
        first = True
        for key, item in self.vocabulary.items():
            if first:
                s += str(item)
                first = False
            else:
                s += f', {item:!s}'
        s += '>'
        return s

    def complete(self, text, state):
        """Command-line completion of the user text in the given state."""
        response = None
        if state == 0:
            # This is the first time for this text, so build a match list.
            origline = readline.get_line_buffer()
            begin = readline.get_begidx()
            end = readline.get_endidx()
            being_completed = origline[begin:end]
            words = origline.split()

            logging.debug('origline=%s', repr(origline))
            logging.debug('begin=%s', begin)
            logging.debug('end=%s', end)
            logging.debug('being_completed=%s', being_completed)
            logging.debug('words=%s', words)

            if not words:
                self.current_candidates = sorted(self.vocabulary.keys())
            else:
                try:
                    if begin == 0:
                        # first word
                        candidates = self.vocabulary.keys()
                    else:
                        # later word
                        first = words[0]
                        candidates = self.vocabulary[first]

                    if being_completed:
                        # match vocabulary with portion of input
                        # being completed
                        self.current_candidates = \
                            [w for w in candidates if w.startswith(being_completed)]
                    else:
                        # matching empty string so use all candidates
                        self.current_candidates = candidates

                    logging.debug('candidates=%s', self.current_candidates)

                except (KeyError, IndexError) as err:
                    logging.error('completion error: %s', err)
                    self.current_candidates = []

        try:
            response = self.current_candidates[state]
        except IndexError:
            response = None
        logging.debug('complete(%s, %s) => %s', repr(text), state, response)
        return response

    def input_generator(self) -> str:
        """Yield user input whenever an entry is confirmed.

           Automatically quits if `terminal_word` is typed.

           :return: The line of user input.
           :rtype: str
        """
        line = ''
        while line != self.terminal_word:
            line = input(self.prompt)
            yield line


if __name__ == '__main__':
    # Register our completer function
    cm = CommandMenu({
        'list': ['files', 'directories'],
        'print': ['byname', 'bysize'],
        'stop': [],
    }, ':> ', 'stop')
    readline.set_completer(cm.complete)

    # Use the tab key for completion
    readline.parse_and_bind('tab: complete')        # Works on Linux
    readline.parse_and_bind('bind ^I rl_complete')  # Works on macOS
    readline.parse_and_bind('set editing-mode vi')

    # Prompt the user for text
    for line in cm.input_generator():
        print(f'Input was: {line}')



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2025 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-enumaction-py
"""Argparse action for using enum as a choice.

Why would you want to do this if there is already the `choice`

Using it is a simple process.

1. To use it, define an enum subclass with your choices.

.. code-block:: python
    class FooEnum(enum.Enum):
        Foo = 'foo'
        Bar = 'bar'

2. Add an argument to the parser and specify a kwarg `type` with the enum
subclass as value and `action` with value `EnumAction`. Optionally, add a
kwarg `use` with value `name` [default] or `value` to use the enum name or
value, respectively, to match the argument to an enum member.

.. code-block:: python

    parser = argparse.ArgumentParser()
    parser.add_argument('--foo', '-f', type=FooEnum, action=EnumAction)
    parser.add_argument('--bar', '-b', type=FooEnum, use='value', action=EnumAction)

3. Finally, call the :py:func`parse_args()` function of the parser.

Inspired by https://stackoverflow.com/a/60750535.

.. versionadded:: 1.7.0
"""

import argparse
import enum
from pprint import pformat


class EnumAction(argparse.Action):
    """Argparse action for handling Enums"""

    def __init__(self, **kwargs):
        """Initialize the instance.

            Keyword arguments:
                - *type* Specify an :py:type:`enum.Enum` subclass. [required]
                - *use* Either "name" or "value". [optional]
        """
        # Get and check the required 'type' argument.
        enum_type = kwargs.pop('type', None)
        if enum_type is None:
            raise ValueError(f'kwarg "type" missing in {self.__class.__name__}')
        if not issubclass(enum_type, enum.Enum):
            raise TypeError(f'kwarg "type" must specify an enum.Enum subclass in {self.__class.__name__}')
        self._enum_type = enum_type

        # Get and check the optional 'use' argument.
        enum_use = kwargs.pop('use', 'name').lower()
        if enum_use not in ('name', 'value'):
            raise ValueError(f'kwarg "use" must be "name" or "value" in {self.__class.__name__}')
        self._use_name = (enum_use == 'name')

        # Generate choices from the given enum.
        if self._use_name:
            kwargs.setdefault('choices', tuple(e.name for e in enum_type))
        else:
            kwargs.setdefault('choices', tuple(e.value for e in enum_type))

        # Perform the rest of the instance initialization.
        super(EnumAction, self).__init__(**kwargs)

    def __call__(self, parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values,
                 option_string:str=None):
        """Perform the action.

        :param argparse.ArgumentParser parser: The parser object containing
             this action.
        :param argparse.Namespace namespace: The namespace object returned by
             :py:func:`parse_args()`.
        :param values: The associated command-line arguments, with any
            type conversions applied. Type conversions are specified with the
            `type` keyword argument to :py:func:`add_argument()`.
        :param str option_string: The option string that was used to invoke
            this action. The option_string argument is optional, and will be
            absent if the action is associated with a positional argument.
        """
        if self._use_name:
            value = self._enum_type[values]
        else:
            value = self._enum_type(values)
        setattr(namespace, self.dest, value)


def example():
    """Example usage."""
    class FooEnum(enum.Enum):
        Foo = 'foo'
        Bar = 'bar'

    parser = argparse.ArgumentParser(exit_on_error=False)
    parser.add_argument('--name', '-n', type=FooEnum, action=EnumAction)
    parser.add_argument('--value', '-v', type=FooEnum, use='value', action=EnumAction)

    # Good case.
    argv = ['--name=Foo', '-v', 'bar']
    print(f'Argv  : {pformat(argv)}')
    args = parser.parse_args(argv)
    print(f'Result: {pformat(args)}')

    # Use value instead of name.
    argv = ['-n', 'foo']
    print(f'Argv  : {pformat(argv)}')
    try:
        parser.parse_args(argv)
    except argparse.ArgumentError as e:
        print(e)

    # Use name instead of value
    argv = ['--value', 'Foo']
    print(f'Argv  : {pformat(argv)}')
    try:
        parser.parse_args(argv)
    except argparse.ArgumentError as e:
        print(e)

    argv = ['-n', 'blat']
    print(f'Argv  : {pformat(argv)}')
    try:
        parser.parse_args(argv)
    except argparse.ArgumentError as e:
        print(e)


if __name__ == '__main__':
    example()



#!/usr/bin/env python3
# vim: fileencoding=utf-8 ts=4
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges. All rights reserved.
# SPDXID: SPDXRef-test-hardwareregister-py
import enum
from typing import Tuple

import pytest

from pymisclib.hardwareregister import (
    Register, RegisterDescription, RegisterFieldBool, RegisterFieldEnum,
    RegisterFieldInt, RegisterValue)


@pytest.mark.parametrize('bit', [x for x in range(0, 64)])
def test_register_field_bool(bit: int):
    fb = RegisterFieldBool('fb', bit)
    assert isinstance(fb, RegisterFieldBool)
    assert fb.name == 'fb'
    assert fb.conversion == bool
    assert fb.first_bit == fb.last_bit
    assert fb.first_bit == bit
    assert fb.mask == (1 << bit)
    assert fb.convert(0) == False
    assert fb.convert(1) == True


def bit_ranges() -> Tuple[int, int]:
    for x in range(0, 64):
        for y in range(x, 64):
            yield x, y

@pytest.mark.parametrize('first_bit, last_bit', [x for x in bit_ranges()])
def test_register_field_int_ok(first_bit: int, last_bit: int):
    f_int = RegisterFieldInt('f_int', first_bit, last_bit)
    assert isinstance(f_int, RegisterFieldInt)
    assert f_int.name == 'f_int'
    assert f_int.conversion == int
    assert f_int.first_bit == first_bit
    assert f_int.last_bit == last_bit
    assert f_int.min_value == 0
    assert f_int.max_value == (1 << (last_bit - first_bit + 1)) - 1
    assert f_int.mask == ((1 << (last_bit + 1)) - 1) & ~((1 << first_bit) - 1)
    assert f_int.convert((1 << first_bit) - last_bit) == (1 << first_bit) - last_bit


def test_register_field_int_bit_range_error():
    with pytest.raises(IndexError):
        RegisterFieldInt('f8', 1, 0)


def test_register_field_int_enum():
    @enum.unique
    class E1(enum.IntEnum):
        zero = 0
        one = 1
        two = 2
        three = 3

    f_ie = RegisterFieldEnum('f_ie', 7, 8, E1)
    assert isinstance(f_ie, RegisterFieldEnum)
    assert f_ie.name == 'f_ie'
    assert f_ie.conversion == E1
    assert f_ie.first_bit == 7
    assert f_ie.last_bit == 8
    assert f_ie.mask == 0x180
    assert f_ie.convert(0) == E1.zero
    assert f_ie.convert(1) == E1.one
    assert f_ie.convert(2) == E1.two
    assert f_ie.convert(3) == E1.three


def test_register_description():
    rd = RegisterDescription(
        32,
        [RegisterFieldBool('b0', 0),
         RegisterFieldInt('i1_17', 1, 17),
         RegisterFieldInt('i18_21', 18, 21),
         # Bits 21..30 are reserved
         RegisterFieldBool('b31', 31)])

def test_register_description_bit_order():
    total_bits = 32
    rd = RegisterDescription(
        total_bits,
        [RegisterFieldBool('b0', 0),
         RegisterFieldInt('i18_21', 18, 21),
         RegisterFieldInt('i1_17', 1, 17),
         # Bits 21..30 are reserved
         RegisterFieldBool('b31', 31)])
    current_bit = -1
    for field in rd.fields_by_start_bit():
        assert current_bit < field.first_bit
        current_bit = field.last_bit
    assert current_bit < total_bits
    current_bit = total_bits
    for field in rd.fields_by_end_bit():
        assert current_bit > field.last_bit
        current_bit = field.first_bit
    assert current_bit >= 0

def test_register():
    rd = RegisterDescription(
        32,
        [RegisterFieldBool('b0', 0),
         RegisterFieldInt('i1_17', 1, 17),
         RegisterFieldInt('i18_21', 18, 21),
         # Bits 21..30 are reserved
         RegisterFieldBool('b31', 31)])
    reg = Register('register1@0x123', 0x123, rd)


def test_register_value():
    @enum.unique
    class E2(enum.IntFlag):
        b0 = 1
        b1 = 2
        b2 = 4

    rd = RegisterDescription(
        32,
        [RegisterFieldBool('b0', 0),
         RegisterFieldInt('i1_17', 1, 17),
         RegisterFieldInt('i18_21', 18, 21),
         RegisterFieldEnum('e22_24', 22, 24, E2),
         # Bits 25..30 are reserved
         RegisterFieldBool('b31', 31)])
    rv = RegisterValue('register1@0', 0, rd, value=0x80fe2345)
    # 31     24      16       8       0
    #          --------        --------
    #  10000000111111100010001101000101
    #  b31       i18 i1_17            b0
    assert rv.value == 0x80fe2345
    assert rv.b0 == True
    assert rv.i1_17 == 0x111a2
    assert rv.i18_21 == 0xf
    assert rv.e22_24 == E2.b0 | E2.b1
    assert rv.b31 == True
    rv.set('b0', False)
    assert rv.value == 0x80fe2344
    rv.set('i1_17', 0x11a2)
    assert rv.value == 0x80fc2344
    rv.set('e22_24', E2.b2)
    assert rv.value == 0x813c2344



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges
# SPDXID: SPDXRef-test-ringbuffer-py

import unittest


from pymisclib.ringbuffer import BufferEmptyError, BufferFullError, RingBuffer

class TestRingBuffer(unittest.TestCase):

    def setUp(self):
        self.capacity = 16
        self.ring_buffer = RingBuffer(self.capacity)

    def test_empty(self):
        """Test behavior of empty ring buffer."""
        self.assertEqual(self.capacity, self.ring_buffer.capacity)
        self.assertEqual(0, self.ring_buffer.size)
        self.assertEqual(0, self.ring_buffer.num_unread)
        with self.assertRaises(BufferEmptyError):
            self.ring_buffer.get()
        unread = self.ring_buffer.get_unread()
        self.assertEqual(0, len(unread))

    def test_full(self):
        """Test behavior of full ring buffer."""
        for i in range(self.capacity):
            self.ring_buffer.add(i)
        self.assertEqual(self.capacity, self.ring_buffer.capacity)
        self.assertEqual(self.capacity, self.ring_buffer.size)
        self.assertEqual(self.capacity, self.ring_buffer.num_unread)
        with self.assertRaises(BufferFullError):
            self.ring_buffer.add(self.capacity + 1)
        unread = self.ring_buffer.get_unread()
        self.assertEqual(self.capacity, len(unread))
        for i in range(self.capacity):
            self.assertEqual(i, unread[i])

    def test_add_one(self):
        """Add one item to ring buffer."""
        self.ring_buffer.add(1)
        self.assertEqual(self.capacity, self.ring_buffer.capacity)
        self.assertEqual(1, self.ring_buffer.size)
        self.assertEqual(1, self.ring_buffer.num_unread)
        unread = self.ring_buffer.get_unread()
        self.assertEqual(1, len(unread))
        self.assertEqual(1, unread[0])

    def test_get_one(self):
        """Get one item from ring buffer."""
        self.ring_buffer.add(123)
        self.assertEqual(123, self.ring_buffer.get())
        self.assertEqual(self.capacity, self.ring_buffer.capacity)
        self.assertEqual(1, self.ring_buffer.size)
        self.assertEqual(0, self.ring_buffer.num_unread)
        unread = self.ring_buffer.get_unread()
        self.assertEqual(0, len(unread))

    def test_add_get(self):
        for i in range(self.capacity * 3 + 1):
            val = str(i)
            self.ring_buffer.add(val)
            self.assertEqual(min(self.capacity, i + 1), self.ring_buffer.size)
            self.assertEqual(1, self.ring_buffer.num_unread)
            self.assertEqual(val, self.ring_buffer.get())
            self.assertEqual(min(self.capacity, i + 1), self.ring_buffer.size)
            self.assertEqual(0, self.ring_buffer.num_unread)
        self.assertEqual(self.capacity, self.ring_buffer.capacity)

    def test_add_overflow(self):
        """Add more items than are read from the ring buffer."""
        for i in range(0, (self.capacity - 2) * 2, 2):
            iteration = int(i / 2) + 1
            self.ring_buffer.add(i)
            self.ring_buffer.add(i + 1)
            val = self.ring_buffer.get()
            self.assertEqual(iteration - 1, val)
            self.assertEqual(min(self.capacity, iteration * 2), self.ring_buffer.size)
            self.assertEqual(iteration, self.ring_buffer.num_unread)
            unread = self.ring_buffer.get_unread()
            self.assertEqual(iteration, len(unread))
            for u in range(i + 1, i + 1 - iteration, -1):
                self.assertEqual(unread.pop(), u)
        i = (self.capacity - 2) * 2
        self.ring_buffer.add(i)
        self.assertEqual(self.capacity - 1, self.ring_buffer.num_unread)
        with self.assertRaises(BufferFullError):
            self.ring_buffer.add(i + 1)

    def test_get_underflow(self):
        """Read more items than are added to the ring buffer."""
        end1 = (self.capacity - 2) * 2
        for i in range(0, end1, 2):
            iteration = int(i / 2) + 1
            self.ring_buffer.add(i)
            self.ring_buffer.add(i + 1)
            val = self.ring_buffer.get()
            self.assertEqual(iteration - 1, val)
            self.assertEqual(min(self.capacity, iteration * 2), self.ring_buffer.size)
            self.assertEqual(iteration, self.ring_buffer.num_unread)
            unread = self.ring_buffer.get_unread()
            self.assertEqual(self.ring_buffer.num_unread, len(unread))
        self.ring_buffer.add(end1)
        self.assertEqual(self.capacity - 1, self.ring_buffer.num_unread)
        num_iterations = self.ring_buffer.num_unread - 1
        for i in range(num_iterations):
            self.ring_buffer.get()
            self.ring_buffer.get()
            self.ring_buffer.add(end1 + i + 1)
            self.assertEqual(num_iterations - i, self.ring_buffer.num_unread)
            unread = self.ring_buffer.get_unread()
            self.assertEqual(self.ring_buffer.num_unread, len(unread))
            for u in range(end1 + i + 1, end1 + i + 1 - len(unread), -1):
                self.assertEqual(u, unread.pop())
        self.assertEqual(1, self.ring_buffer.num_unread)
        self.ring_buffer.get()
        self.assertEqual(0, self.ring_buffer.num_unread)
        with self.assertRaises(BufferEmptyError):
            self.ring_buffer.get()


if __name__ == '__main__':
    unittest.main()
